﻿using System.Text;
using System.Net.Http.Headers;

using Newtonsoft.Json;

namespace TestProject;

/// <summary>
/// API 클라이언트
/// </summary>
public class APIClient : IDisposable
{
    //////////////////////////////////////////////////////////////////////////////////////////////////// Event
    ////////////////////////////////////////////////////////////////////////////////////////// Public

    #region 메시지 도착시 이벤트 - MessageArrived

    /// <summary>
    /// 메시지 도착시 이벤트
    /// </summary>
    public event EventHandler<StreamingMessage> MessageArrived;

    #endregion

    //////////////////////////////////////////////////////////////////////////////////////////////////// Field
    ////////////////////////////////////////////////////////////////////////////////////////// Private

    #region Field

    /// <summary>
    /// 리소스 해제 여부
    /// </summary>
    private bool disposed = false;

    /// <summary>
    /// BASE URL
    /// </summary>
    private const string BASE_URL = "https://clovastudio.stream.ntruss.com/testapp/v1/chat-completions/HCX-DASH-001";

    /// <summary>
    /// 클로바 스튜디오 API 키
    /// </summary>
    private string clovaStudioAPIKey;

    /// <summary>
    /// API 게이트웨이 API 키
    /// </summary>
    private string apiGatewayAPIKey;

    /// <summary>
    /// HTTP 클라이언트
    /// </summary>
    private HttpClient client;

    /// <summary>
    /// 품질을 갖는 스트림 미디어 타입 헤더 값
    /// </summary>
    private MediaTypeWithQualityHeaderValue streamMediaTypeWithQualityHeaderValue;

    #endregion

    //////////////////////////////////////////////////////////////////////////////////////////////////// Property
    ////////////////////////////////////////////////////////////////////////////////////////// Public

    #region 클로바 스튜디오 API 키 - ClovaStudioAPIKey

    /// <summary>
    /// 클로바 스튜디오 API 키
    /// </summary>
    public string ClovaStudioAPIKey
    {
        get
        {
            return this.clovaStudioAPIKey;
        }
    }

    #endregion
    #region API 게이트웨이 API 키 - APIGatewayAPIKey

    /// <summary>
    /// API 게이트웨이 API 키
    /// </summary>
    public string APIGatewayAPIKey
    {
        get
        {
            return this.apiGatewayAPIKey;
        }
    }

    #endregion
    #region HTTP 클라이언트 - Client

    /// <summary>
    /// HTTP 클라이언트
    /// </summary>
    public HttpClient Client
    {
        get
        {
            return this.client;
        }
    }

    #endregion

    //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
    ////////////////////////////////////////////////////////////////////////////////////////// Public

    #region API 클라이언트 - APIClient(clovaStudioAPIKey, apiGatewayAPIKey)

    /// <summary>
    /// API 클라이언트
    /// </summary>
    /// <param name="clovaStudioAPIKey">클로바 스튜디오 API 키</param>
    /// <param name="apiGatewayAPIKey">API 게이트웨이 API 키</param>
    public APIClient(string clovaStudioAPIKey, string apiGatewayAPIKey)
    {
        this.clovaStudioAPIKey = clovaStudioAPIKey;
        this.apiGatewayAPIKey  = apiGatewayAPIKey;

        this.client = new HttpClient();

        this.client.DefaultRequestHeaders.Add("X-NCP-CLOVASTUDIO-API-KEY", this.clovaStudioAPIKey);
        this.client.DefaultRequestHeaders.Add("X-NCP-APIGW-API-KEY"      , this.apiGatewayAPIKey );

        this.streamMediaTypeWithQualityHeaderValue = new MediaTypeWithQualityHeaderValue("text/event-stream");
    }

    #endregion

    //////////////////////////////////////////////////////////////////////////////////////////////////// Method
    ////////////////////////////////////////////////////////////////////////////////////////// Public

    #region 채팅 요청하기 - RequestChat(requestMessage)

    /// <summary>
    /// 채팅 요청하기
    /// </summary>
    /// <param name="requestMessage">요청 메시지</param>
    /// <returns>응답 메시지 태스크</returns>
    public async Task<ResponseMessage> RequestChat(RequestMessage requestMessage)
    {
        if(this.client.DefaultRequestHeaders.Accept.Contains(this.streamMediaTypeWithQualityHeaderValue))
        {
            this.client.DefaultRequestHeaders.Accept.Remove(this.streamMediaTypeWithQualityHeaderValue);
        }

        if(this.client.DefaultRequestHeaders.Contains("X-NCP-CLOVASTUDIO-REQUEST-ID"))
        {
            this.client.DefaultRequestHeaders.Remove("X-NCP-CLOVASTUDIO-REQUEST-ID");
        }

        string clovaStudioRequestID = Guid.NewGuid().ToString();

        this.client.DefaultRequestHeaders.Add("X-NCP-CLOVASTUDIO-REQUEST-ID", clovaStudioRequestID);

        string requestJSON = JsonConvert.SerializeObject(requestMessage);

        StringContent content = new StringContent(requestJSON, Encoding.UTF8, "application/json");

        HttpResponseMessage responseMessage = await this.client.PostAsync(BASE_URL, content);

        responseMessage.EnsureSuccessStatusCode();

        string responseString = await responseMessage.Content.ReadAsStringAsync();

        return JsonConvert.DeserializeObject<ResponseMessage>(responseString);
    }

    #endregion
    #region 채팅 요청하기 (스트림) - RequestChatStream(requestMessage)

    /// <summary>
    /// 채팅 요청하기 (스트림)
    /// </summary>
    /// <param name="requestMessage">요청 메시지</param>
    /// <returns>스트리밍 메시지 리스트 태스크</returns>
    public async Task<List<StreamingMessage>> RequestChatStream(RequestMessage requestMessage)
    {
        if(!this.client.DefaultRequestHeaders.Accept.Contains(this.streamMediaTypeWithQualityHeaderValue))
        {
            this.client.DefaultRequestHeaders.Accept.Add(this.streamMediaTypeWithQualityHeaderValue);
        }

        if(this.client.DefaultRequestHeaders.Contains("X-NCP-CLOVASTUDIO-REQUEST-ID"))
        {
            this.client.DefaultRequestHeaders.Remove("X-NCP-CLOVASTUDIO-REQUEST-ID");
        }

        string clovaStudioRequestID = Guid.NewGuid().ToString();

        this.client.DefaultRequestHeaders.Add("X-NCP-CLOVASTUDIO-REQUEST-ID", clovaStudioRequestID);

        string requestJSON = JsonConvert.SerializeObject(requestMessage);

        StringContent content = new StringContent(requestJSON, Encoding.UTF8, "application/json");

        using var request = new HttpRequestMessage(HttpMethod.Post, BASE_URL);

        request.Content = content;

        using HttpResponseMessage responseMessage = await client.SendAsync(request, HttpCompletionOption.ResponseHeadersRead);

        responseMessage.EnsureSuccessStatusCode();

        using Stream stream = await responseMessage.Content.ReadAsStreamAsync();

        using StreamReader reader = new StreamReader(stream);

        List<StreamingMessage> streamingMessageList = new List<StreamingMessage>();

        StreamingMessage currentStreamingMessage = null;

        while(!reader.EndOfStream)
        {
            string line = await reader.ReadLineAsync();

            if(string.IsNullOrWhiteSpace(line))
            {
                continue;
            }

            if(currentStreamingMessage == null)
            {
                if(line.StartsWith("id:"))
                {
                    currentStreamingMessage = new StreamingMessage();

                    currentStreamingMessage.ID = line.Substring(3).Trim();

                    continue;
                }

                if(line.StartsWith("event:"))
                {
                    continue;
                }

                if(line.StartsWith("data:"))
                {
                    continue;
                }
            }
            else
            {
                if(line.StartsWith("id:"))
                {
                    continue;
                }

                if(line.StartsWith("event:"))
                {
                    currentStreamingMessage.Event = line.Substring(6).Trim();
                }

                if(line.StartsWith("data:"))
                {
                    string dataJSON = line.Substring(5).Trim();

                    if(currentStreamingMessage.Event == "token")
                    {
                        StreamingDataToken token = JsonConvert.DeserializeObject<StreamingDataToken>(dataJSON);

                        currentStreamingMessage.Data = token;
                    }
                    else if(currentStreamingMessage.Event == "result")
                    {
                        StreamingDataResult result = JsonConvert.DeserializeObject<StreamingDataResult>(dataJSON);

                        currentStreamingMessage.Data = result;
                    }
                    else if(currentStreamingMessage.Event == "signal")
                    {
                        StreamingDataSignal signal = JsonConvert.DeserializeObject<StreamingDataSignal>(dataJSON);

                        currentStreamingMessage.Data = signal;
                    }

                    streamingMessageList.Add(currentStreamingMessage);

                    MessageArrived?.Invoke(this, currentStreamingMessage);

                    currentStreamingMessage = null;

                    continue;
                }
            }
        }

        return streamingMessageList;
    }

    #endregion

    #region 리소스 해제하기 - Dispose()

    /// <summary>
    /// 리소스 해제하기
    /// </summary>
    public void Dispose()
    {
        Dispose(true);

        GC.SuppressFinalize(this);
    }

    #endregion

    ////////////////////////////////////////////////////////////////////////////////////////// Protected

    #region 리소스 해제하기 - Dispose(disposing)

    /// <summary>
    /// 리소스 해제하기
    /// </summary>
    /// <param name="disposing">리소스 해제 여부</param>
    protected virtual void Dispose(bool disposing)
    {
        if(!this.disposed)
        {
            if(disposing)
            {
                if(this.client != null)
                {
                    this.client.Dispose();
                }

                this.client = null;

                this.disposed = true;
            }
        }
    }

    #endregion
}